
#include <stdio.h>

#include <windows.h>
#include <bfd.h>

#include "bt.h"

typedef struct _Sw_Find_Data Sw_Find_Data;

struct _Sw_Find_Data
{
  char     *function;
  asymbol **symbol_table;
  bfd_vma   counter;
};

struct _Sw
{
  bfd *fd;
  asymbol **symbol_table;
};

Sw *
sw_init(void)
{
  char filename[MAX_PATH];
  Sw *sw;
  char **formats = NULL;
  unsigned int dummy = 0;

  if (!GetModuleFileName(NULL, filename, sizeof(filename)))
    return NULL;

  sw = (Sw *)calloc(1, sizeof(Sw));
  if (!sw)
    return NULL;

  bfd_init();

  sw->fd = bfd_openr(filename, 0);
  if (!sw->fd)
    goto free_sw;

  if (!bfd_check_format(sw->fd, bfd_object))
    goto close_fd;
  if (!bfd_check_format_matches(sw->fd, bfd_object, &formats))
    goto close_fd;
  if (!(bfd_get_file_flags(sw->fd) & HAS_SYMS))
    {
      free(formats);
      goto close_fd;
    }
  free(formats);

  if ((bfd_read_minisymbols(sw->fd, FALSE, (void **)&sw->symbol_table, &dummy) == 0) &&
      (bfd_read_minisymbols(sw->fd, TRUE, (void **)&sw->symbol_table, &dummy) < 0))
    goto close_fd;

  return sw;

 close_fd:
  if (sw->symbol_table)
    free(sw->symbol_table);
  bfd_close(sw->fd);
 free_sw:
  free(sw);

  return NULL;
}

void
sw_shutdown(Sw *sw)
{
  if (!sw)
    return;

  free(sw->symbol_table);
  bfd_close(sw->fd);
  free(sw);
}
static void
sw_find_function_name_in_section(bfd      *abfd,
                                 asection *sec,
                                 void     *obj)
{
  Sw_Find_Data *data;
  bfd_vma       vma;
  const char   *func = NULL;
  const char   *file = NULL;
  unsigned      line = 0;

  if (!sec)
    {
      printf("no section\n");
      return;
    }

  data = (Sw_Find_Data *)obj;
  if (data->function && (*data->function != '\0'))
    {
      /* printf("function already found\n"); */
      return;
    }

  if (!(bfd_get_section_flags(abfd, sec) & SEC_ALLOC))
    return;

  vma = bfd_get_section_vma(abfd, sec);
  if ((data->counter < vma) ||
      ((vma + bfd_get_section_size(sec)) <= data->counter))
    return;

  if (bfd_find_nearest_line(abfd, sec,
                            data->symbol_table,
                            data->counter - vma,
                            &file, &func, &line))
    {
      printf("%s (%s) : %d\n", file, func, line);
      if (func) data->function = strdup(func);
    }
}

void
sw_frames_get(Sw *sw)
{
  void        *frames[50];
  unsigned short frames_nbr;
  unsigned int i;

  frames_nbr = CaptureStackBackTrace(0, 50, frames, NULL);
  if (frames_nbr == 0)
    {
      printf("err %ld\n", GetLastError());
    }

  for (i = 0; i < frames_nbr; i++)
    {
      Sw_Find_Data data;

      data.function = NULL;
      data.symbol_table = sw->symbol_table;
      /* we substract 1 because */
      /* the back-trace address collected is the return-address of the call. */
      /* So this location might be pointing already to next line.*/
      data.counter = ((char *)frames[i] - 1);
      bfd_map_over_sections(sw->fd, &sw_find_function_name_in_section, &data);
    }
}
